from django.db.models import Q

from django_forest.resources.utils.queryset.filters.date.factory import ConditionFactory
from django_forest.resources.utils.queryset.filters.date.conditions import (
    BaseCondition,
    Operator,
    Condition,
    Conditions
)
from django_forest.utils.date import get_now_aware_datetime

class DjangoFieldConditionSerializer:

    OPERATOR = {
        Operator.lt: '__lt',
        Operator.lte: '__lte',
        Operator.gt: '__gt',
        Operator.gte: '__gte',
    }

    def __init__(self, field: str, *args, **kwargs):
        self.field = field

    def serialize_condition(self, condition: Condition):
        serialized_operator = self.OPERATOR[condition.operator]
        return (
            f"{self.field}{serialized_operator}",
            condition.value
        )

    def serialize_many_conditions(self, conditions: Conditions):
        return map(self.serialize_condition, conditions.conditions)

    def serialize(self, condition: BaseCondition):
        serialized = []
        if condition.HAS_MANY:
            serialized = self.serialize_many_conditions(condition)
        else:
            serialized = [self.serialize_condition(condition)]

        return Q(**dict(serialized))

class DatesMixin:
    # Notice, the previous keyword is related to value charts
    previous = False

    def handle_date_operator(self, operator, field, value, tz):
        factory = ConditionFactory(
            get_now_aware_datetime(tz),
        )
        condition = factory.build(
            operator,
            value,
            offset=(1 if self.previous else 0)
        )
        serializer = DjangoFieldConditionSerializer(field)
        return serializer.serialize(condition)



# Generated by Django 3.2.4 on 2021-06-15 09:42

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='question',
            name='pub_date',
            field=models.DateTimeField(default=datetime.datetime(2020, 5, 17, 0, 0), verbose_name='date published'),
        ),
        migrations.AlterField(
            model_name='question',
            name='question_text',
            field=models.CharField(default='how are you?', max_length=200),
        ),
    ]



import logging
from unittest import mock

from django.test import TestCase
from django.urls import reverse
from django.utils.http import urlencode

from django_forest.authentication.oidc.client_manager import OidcClientManager
from django_forest.tests.authentication.oidc.test_configuration_retriever import mocked_config
from django_forest.tests.authentication.oidc.test_dynamic_client_registrator import mocked_client_credentials
from django_forest.tests.utils.test_forest_api_requester import mocked_requests

mocked_token_response = {
    'id_token': 'eyJraWQiOiI2eWthczh5SVMxdHM3dmZZY3JJRTd3aG1Wc2hwbHVKLUlDbU1pRWszYWJ3IiwiYWxnIjoiUlMyNTYifQ.eyJzdWIiOjUsImF1ZCI6ImV5SnJhV1FpT2lJMmVXdGhjemg1U1ZNeGRITTNkbVpaWTNKSlJUZDNhRzFXYzJod2JIVktMVWxEYlUxcFJXc3pZV0ozSWl3aVlXeG5Jam9pVWxNeU5UWWlmUS5leUowYjJ0bGJsOWxibVJ3YjJsdWRGOWhkWFJvWDIxbGRHaHZaQ0k2SW01dmJtVWlMQ0p5WldScGNtVmpkRjkxY21seklqcGJJbWgwZEhBNkx5OXNiMk5oYkdodmMzUTZPREF3TUM5bWIzSmxjM1F2WVhWMGFHVnVkR2xqWVhScGIyNHZZMkZzYkdKaFkyc2lYU3dpY21WbmFYTjBjbUYwYVc5dVgyVnVaSEJ2YVc1MElqb2lhSFIwY0hNNkx5OWhjR2t1WkdWMlpXeHZjRzFsYm5RdVptOXlaWE4wWVdSdGFXNHVZMjl0TDI5cFpHTXZjbVZuSWl3aVlYQndiR2xqWVhScGIyNWZkSGx3WlNJNkluZGxZaUlzSW1keVlXNTBYM1I1Y0dWeklqcGJJbUYxZEdodmNtbDZZWFJwYjI1ZlkyOWtaU0pkTENKeVpYTndiMjV6WlY5MGVYQmxjeUk2V3lKamIyUmxJbDBzSW1WdWRtbHliMjV0Wlc1MFgybGtJam94TURFc0ltbHpjeUk2SWtaUFVrVlRWRjlCVlZSSVJVNVVTVU5CVkVsUFRsOVRXVk5VUlUwaUxDSnBZWFFpT2pFMk1qTTBNamM1TlRWOS5QX21DOGVDcjM2UnFSeWU4WUk1elVTR3VxNEdtcEZmOFJMdnpRbW5WTWsySFhxelFUb1Bhb19SQXl2R3BZQzRKVWpGbXVUZEdJNlNTZzlKVkdJek5vZDJBWEZBeHA4cW1LOGF4X1RCS1VsZ19GX216cnpPeEVPYWlGa2lDUUFoYXZsQmhOQ3g1Nl9DUkJteXRfNVV2NkwtRTRfZXpHMEExa2FjUGZPUFphQk5TajhGcUNSeHJzblZaVnE2MFJtWTd0UmwwRFg1MFpVdVNFdmgwenpaTHdheUJ1VFlneExMZGc2RXZSNTJkc0pIbHYwS2dObG5ISndLQjc0Y1NBbDlDTTNQVkh5bTlFZzNWM0JfQmFBYnUwZUlLbXF0SjdfMWgxdzMtWXNINGpBQy1jelFOZ3pnTzc5SWlZMnMwdFRSRWU0dnYtUC1fX0JKOGxJeUZ4Y2d6dVEiLCJpc3MiOiJodHRwczovL2FwaS5kZXZlbG9wbWVudC5mb3Jlc3RhZG1pbi5jb20iLCJpYXQiOjE2MjM0Mjc5NjksImV4cCI6MTYyMzQzMTU2OX0.tqBAbP41a9U-XpX8Q_w3dRgVDqDtInTAGlzpwUvrS-49Mp7YSkB_eGypuYR_Kk7VGfjibCZflR6EXvbzg_5CrOKcpr-b44xi1k9zQntBtc7YPNJV30O0mPZW3ZAAWiZww-8LA8F2YyDK7l7VyoPO6oygTRtM6eJH91HQV9dE2sPILwm5jO-OutuwA0ZXGbUDL7OJ9if-V26IAKlurfqaUtL__RRs9s4uvRgvJzGEWGdImQYjzGtHDwEAQnMr8g3fwQWOlHa4xNbh3GWzkCiR_Tkh8d1zmXEWqx60Bv_IW9m3Db8ZR4_h8vG_7rbGlhVu1YeQjJ6jwk3kgeNq-KYIVA',
    'access_token': 'eyJraWQiOiI2eWthczh5SVMxdHM3dmZZY3JJRTd3aG1Wc2hwbHVKLUlDbU1pRWszYWJ3IiwiYWxnIjoiUlMyNTYifQ.eyJkYXRhIjp7ImRhdGEiOnsidHlwZSI6InVzZXJzIiwiaWQiOiI1IiwiYXR0cmlidXRlcyI6eyJmaXJzdF9uYW1lIjoiR3VpbGxhdW1lIiwibGFzdF9uYW1lIjoiQ2lzY28iLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIn19fSwic2NvcGUiOiJvcGVuaWQgZW1haWwgcHJvZmlsZSIsImF1ZCI6IkZPUkVTVF9VU0VSUyIsImlzcyI6IkZPUkVTVF9BVVRIRU5USUNBVElPTl9TWVNURU0iLCJpYXQiOjE2MjM0Mjc5NjksImV4cCI6MTYyMzQzMTU2OX0.moA3J5u0JmbEhXgrWFP-w4dyIQdoSGIzKIK7fifZctF00EA_B_aeWnQWSfUY-Bc9czSm6wg9bzNQ5d1brhcvMzrnG4OSYm3g40Nq0WQktysIvnyIuVPgN_iUjVu4cai71ndwhHOKBSHsJRZ7-nQdluk9nQQ23BH6QmMZsRJnmtolzi8tHp3Rfozb8lO9cwY1bZzaAqzrj0MQxMgr1j7eReRGl4-v2W8WCMcaFdtcVMC6Rr98rhnIrbCJtWedJIqrkkpV_vFbCOJXMwFKZ3-VBu13MGvGu033BlbAOuw-cBHsXi3AKf3PkJ0aFXXrcFf_3AyEWqfoSE8uyjF66gCTFA',
    'expires_in': 36000000000,
    'scope': 'openid email profile',
    'token_type': 'Bearer'
}

mocked_jwks = {
    'keys': [
        {
            'e': 'AQAB',
            'n': 'z6Jiw8wmc5PEnYMoRNogKkqLi-nmoNY5dlYR0vb5Mhs5uLjgiZ0hkDKSTvTVHvLR6Yy9r7q7KI7A1cyhnhffK8J30plETLFbve5lVfGgYfEtYdMvJPmQznrMnfAJp6w_WDK8rqfR6N2hgFAQmkekf-ROC6Mj7XKtA4hDIGLsG_sv7ai_NcHNNVe5DkbSn3NjHpw0iIaPpjrAUmm9Npr6K45UreHd_6TCd9ddbUEgCFe6WtpboGyPOC18Sa6tSp-34-NCFefA7krnbYBBPRCy3szGxlTcx_JsljC_amRCNFraeBpswVzE5-Hq7-bpCX9crmOTZLWp8PG5d70AfixCEQ',
            'kty': 'RSA',
            'kid': '6ykas8yIS1ts7vfYcrIE7whmVshpluJ-ICmMiEk3abw',
            'alg': 'RS256',
            'use': 'sig'
        }
    ]
}

mocked_user = {
    'data': {
        'type': 'users',
        'id': '5',
        'attributes': {
            'first_name': 'Guillaume', 'last_name': 'Cisco',
            'email': 'guillaumec@forestadmin.com',
            'teams': ['Operations'],
            'two_factor_authentication_enabled': False,
            'two_factor_authentication_active': False,
            'two_factor_authentication_secret': None
        }
    }
}


def mocked_requests_get(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        return mocked_requests(mocked_user, 200)


def mocked_requests_get_error_authorization(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        raise Exception('Authorization error')


def mocked_requests_get_not_found(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        return mocked_requests({}, 404)


def mocked_requests_get_422(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        return mocked_requests({}, 422)


def mocked_requests_get_bad_response_2fa(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        return mocked_requests({'errors': [{'name': 'TwoFactorAuthenticationRequiredForbiddenError'}]}, 500)


def mocked_requests_get_bad_response(value, **args):
    if value == 'https://api.development.forestadmin.com/oidc/jwks':
        return mocked_requests(mocked_jwks, 200)
    elif value == 'https://api.test.forestadmin.com/liana/v2/renderings/1/authorization':
        return mocked_requests({'errors': [{'name': 'error'}]}, 500)


class AuthenticationCallbackViewTests(TestCase):

    def setUp(self):
        self.retrieve_patcher = mock.patch('django_forest.authentication.oidc.client_manager.retrieve',
                                           return_value=mocked_config)
        self.register_patcher = mock.patch('django_forest.authentication.oidc.client_manager.register',
                                           return_value=mocked_client_credentials)
        self.mocked_retrieve = self.retrieve_patcher.start()
        self.mocked_register = self.register_patcher.start()
        callback_url = 'http://localhost:8000/authentication/callback'
        self.oidc_client = OidcClientManager.get_client_for_callback_url(callback_url)
        self.oidc_client.redirect_uris = ['http://localhost:8000/forest/authentication/callback']

        url = reverse('django_forest:authentication:callback')
        query = {
            'code': 'eslqHqk8Luo_3CIf5SanmXBpq_7ytlTV8HgoVFNPwUvmWKDiwnf9XV6Bo04zRon8',
            'state': '{"renderingId": 1}'
        }
        self.url = f'{url}?{urlencode(query)}'

    def tearDown(self):
        OidcClientManager.client = None
        self.oidc_client = None
        self.retrieve_patcher.stop()
        self.register_patcher.stop()

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get)
    def test_get(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        r = response.json()
        self.assertTrue('token' in r)
        self.assertTrue('tokenData' in r)
        self.assertEqual(r['tokenData']['id'], '5')
        self.assertEqual(r['tokenData']['email'], 'guillaumec@forestadmin.com')
        self.assertEqual(r['tokenData']['first_name'], 'Guillaume')
        self.assertEqual(r['tokenData']['last_name'], 'Cisco')
        self.assertEqual(r['tokenData']['team'], 'Operations')
        self.assertEqual(r['tokenData']['rendering_id'], 1)

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get_error_authorization)
    def test_get_error(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        with self.assertLogs() as cm:
            r = self.client.get(self.url)
            self.assertEqual(r.status_code, 500)
            r = r.json()
            self.assertEqual(r, {
                'errors': [
                    {
                        'status': 500,
                        'detail': 'Cannot reach Forest API at https://api.test.forestadmin.com/liana/v2/renderings/1/authorization, it seems to be down right now.'
                    }
                ]
            })

            self.assertEqual(cm.output, [
                'ERROR:django_forest.authentication.views.callback:Authorization error: Cannot reach Forest API at https://api.test.forestadmin.com/liana/v2/renderings/1/authorization, it seems to be down right now.',
                'ERROR:django.request:Internal Server Error: /forest/authentication/callback'
            ])

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get_not_found)
    def test_get_not_found(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        r = self.client.get(self.url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Cannot retrieve the data from the Forest server. Can you check that you properly copied the Forest envSecret in your settings?'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get_422)
    def test_get_422(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        r = self.client.get(self.url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Cannot retrieve the project you\'re trying to unlock. The envSecret and renderingId seems to be missing or inconsistent.'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get_bad_response_2fa)
    def test_get_bad_response_2fa(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        r = self.client.get(self.url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Two factor authentication required'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get_bad_response)
    def test_get_bad_response(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        r = self.client.get(self.url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Error while authorizing the user on Forest Admin'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get)
    def test_get_state_missing(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        url = reverse('django_forest:authentication:callback')
        query = {
            'code': 'eslqHqk8Luo_3CIf5SanmXBpq_7ytlTV8HgoVFNPwUvmWKDiwnf9XV6Bo04zRon8',
        }
        url = f'{url}?{urlencode(query)}'
        r = self.client.get(url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Invalid response from the authentication server: the state parameter is missing'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get)
    def test_get_no_rendering_id(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        url = reverse('django_forest:authentication:callback')
        query = {
            'code': 'eslqHqk8Luo_3CIf5SanmXBpq_7ytlTV8HgoVFNPwUvmWKDiwnf9XV6Bo04zRon8',
            'state': '{"foo": 1}'
        }
        url = f'{url}?{urlencode(query)}'
        r = self.client.get(url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Invalid response from the authentication server: the state does not contain a renderingId'
                }
            ]
        })

    @mock.patch('oic.utils.time_util.utc_time_sans_frac', return_value=1623431559)
    @mock.patch('requests.request', return_value=mocked_requests(mocked_token_response, 200))
    @mock.patch('requests.get', side_effect=mocked_requests_get)
    def test_get_invalid_state(self, mocked_requests_get, mocked_requests_request, mocked_utc_time_sans_frac):
        url = reverse('django_forest:authentication:callback')
        query = {
            'code': 'eslqHqk8Luo_3CIf5SanmXBpq_7ytlTV8HgoVFNPwUvmWKDiwnf9XV6Bo04zRon8',
            'state': '{"renderingId": error}'
        }
        url = f'{url}?{urlencode(query)}'
        r = self.client.get(url)
        self.assertEqual(r.status_code, 500)
        r = r.json()
        self.assertEqual(r, {
            'errors': [
                {
                    'status': 500,
                    'detail': 'Invalid response from the authentication server: the state parameter is not at the right format'
                }
            ]
        })


import json
import logging
from datetime import timedelta, datetime
from jose import jwt
from oic.oauth2 import AuthorizationResponse
from django.http import JsonResponse
from django.views.generic import View

from django_forest.authentication.oidc.client_manager import OidcClientManager
from django_forest.authentication.utils import get_callback_url
from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.forest_setting import get_forest_setting


logger = logging.getLogger(__name__)


# Based on https://pyoidc.readthedocs.io/en/latest/examples/rp.html
class CallbackView(View):

    def _expiration_in_seconds(self):
        return datetime.timestamp(datetime.utcnow() + timedelta(hours=1))

    def _handle_2fa_error(self, response):
        body = response.json()
        if 'errors' in body and isinstance(body['errors'], list) and len(body['errors']) > 0:
            server_error = body['errors'][0]
            if 'name' in server_error and server_error['name'] == 'TwoFactorAuthenticationRequiredForbiddenError':
                raise Exception(MESSAGES['SERVER_TRANSACTION']['TWO_FACTOR_AUTHENTICATION_REQUIRED'])

    def _handle_authent_response(self, response):
        if response.status_code == 200:
            body = response.json()
            user = body['data']['attributes']
            user['id'] = body['data']['id']
            return user
        elif response.status_code == 404:
            raise Exception(MESSAGES['SERVER_TRANSACTION']['SECRET_NOT_FOUND'])
        elif response.status_code == 422:
            raise Exception(MESSAGES['SERVER_TRANSACTION']['SECRET_AND_RENDERINGID_INCONSISTENT'])
        else:
            self._handle_2fa_error(response)

        raise Exception(MESSAGES['SERVER_TRANSACTION']['AUTHORIZATION'])

    def _authenticate(self, rendering_id, auth_data):
        route = f'/liana/v2/renderings/{rendering_id}/authorization'
        headers = {'forest-token': auth_data['forest_token']}

        try:
            response = ForestApiRequester.get(route, headers=headers)
        except Exception as e:
            logger.error(f'Authorization error: {str(e)}')
            raise e
        else:
            return self._handle_authent_response(response)

    def _verify_code_and_generate_token(self, redirect_url, request):
        if 'state' not in request.GET:
            raise Exception(MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_MISSING'])
        if 'renderingId' not in request.GET['state']:
            raise Exception(MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_RENDERING_ID'])

        client = OidcClientManager.get_client_for_callback_url(redirect_url)
        aresp = client.parse_response(AuthorizationResponse,
                                      info=request.get_full_path_info(),
                                      sformat='urlencoded',
                                      state=request.GET['state'],
                                      scope=['openid', 'email', 'profile'])
        resp = client.do_access_token_request(state=request.GET['state'],
                                              scope=['openid', 'email', 'profile'],
                                              authn_method='',
                                              request_args={'code': aresp['code']},
                                              verify=False)
        try:
            rendering_id = json.loads(request.GET['state'])['renderingId']
        except Exception:
            raise Exception(MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_FORMAT'])
        else:
            user = self._authenticate(rendering_id, {'forest_token': resp['access_token']})

            auth_secret = get_forest_setting('FOREST_AUTH_SECRET')
            return jwt.encode({
                'id': user['id'],
                'email': user['email'],
                'first_name': user['first_name'],
                'last_name': user['last_name'],
                'team': user['teams'][0],
                'rendering_id': rendering_id,
                'exp': self._expiration_in_seconds()
            },
                auth_secret,
                algorithm='HS256')

    def get(self, request, *args, **kwargs):
        try:
            callback_url = get_callback_url()
            token = self._verify_code_and_generate_token(callback_url, request)

            auth_secret = get_forest_setting('FOREST_AUTH_SECRET')
            result = {
                'token': token,
                'tokenData': jwt.decode(token, auth_secret, algorithms=['HS256'])
            }
        except Exception as error:
            return JsonResponse({'errors': [{'status': 500, 'detail': str(error)}]}, status=500)
        else:
            return JsonResponse(result)



from django.db import connection


def get_row(query):
    with connection.cursor() as cursor:
        cursor.execute(query)
        if not len(cursor.description) == 1:
            raise Exception("The result column must be named 'value'")
        value = cursor.description[0].name
        if value != 'value':
            raise Exception(f"The result column must be named 'value' instead of '{value}'.")
        row = cursor.fetchone()

    return row[0]


def check_query(description, column_1, column_2):
    if not len(description) == 2:
        raise Exception(f"The result columns must be named '{column_1}' and '{column_2}'.")
    first = description[0].name
    second = description[1].name
    if first != column_1 or second != column_2:
        err_msg = f"The result columns must be named '{column_1}', '{column_2}' instead of '{first}', '{second}'."
        raise Exception(err_msg)


def execute_query(query, column_1, column_2, one=False):
    with connection.cursor() as cursor:
        cursor.execute(query)
        check_query(cursor.description, column_1, column_2)
        if one:
            res = cursor.fetchone()
        else:
            res = cursor.fetchall()

    return res



import copy
from datetime import datetime
from unittest import mock

import pytest
import pytz
from django.test import TransactionTestCase
from django.urls import reverse
from freezegun import freeze_time

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.utils.test_forest_api_requester import mocked_requests
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager
from django_forest.utils.date import get_timezone

mocked_scope = {
    'tests_question': {
        'scope': {
            'filter': {
                'aggregator': 'and',
                'conditions': [
                    {
                        'field': 'question_text',
                        'operator': 'contains',
                        'value': 'color'
                    },
                ]
            },
            'dynamicScopesValues': {}
        }
    }
}

mocked_scope_dynamic_value = {
    'tests_question': {
        'scope': {
            'filter': {
                'aggregator': 'or',
                'conditions': [
                    {
                        'field': 'question_text',
                        'operator': 'contains',
                        'value': 'color'
                    },
                    {
                        'field': 'question_text',
                        'operator': 'contains',
                        'value': '$currentUser.name'
                    }
                ]
            },
            'dynamicScopesValues': {
                'users': {
                    '1': {
                        '$currentUser.name': 'singer'
                    }
                }
            }
        }
    }
}


def mocked_requests_scope(value, *args):
    def m(url, **kwargs):
        if url == 'https://api.test.forestadmin.com/liana/scopes':
            return mocked_requests(value['scope']['data'], value['scope']['status'])

    return m


class ResourceListScopeViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.reverse_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_choice'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_scope_cached(self, mocked_decode):
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {'topic': {'data': None,
                                                'links': {'related': '/forest/tests_question/1/relationships/topic'}}},
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('requests.get', side_effect=mocked_requests_scope({
        'scope': {
            'data': mocked_scope,
            'status': 200
        }
    }))
    def test_scope_simple(self, mocked_requests, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {'topic': {'data': None,
                                                'links': {'related': '/forest/tests_question/1/relationships/topic'}}},
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1, 'name': 'singer'})
    @mock.patch('requests.get', side_effect=mocked_requests_scope({
        'scope': {
            'data': mocked_scope_dynamic_value,
            'status': 200
        }
    }))
    def test_scope_aggregator_dynamic_values(self, mocked_requests, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {'topic': {'data': None,
                                                'links': {'related': '/forest/tests_question/1/relationships/topic'}}},
                },
                {
                    'type': 'tests_question',
                    'id': 3,
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {'topic': {'data': None,
                                                'links': {'related': '/forest/tests_question/3/relationships/topic'}}},
                },
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('requests.get', side_effect=mocked_requests_scope({
        'scope': {
            'data': {},
            'status': 400
        }
    }))
    def test_scope_no_response(self, mocked_requests, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Unable to fetch scopes'
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_scope_bad_token(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Missing required rendering_id'
                }
            ]
        })



class PaginationMixin:
    def get_pagination(self, params, queryset):
        if 'page[number]' in params and 'page[size]' in params:
            page_number = int(params['page[number]'])
            page_size = int(params['page[size]'])

            _from = (page_number - 1) * page_size
            _to = page_number * page_size

            return queryset[_from:_to]

        return queryset



import sys

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata


def get_app_version():
    version = '0.0.0'
    try:
        version = metadata.version('django-forestadmin')
    except Exception:
        pass
    finally:
        return version



import json
from datetime import datetime
from unittest import mock

import pytz
from django.test import TestCase
from django.urls import reverse

from django_forest.tests.resources.views.list.test_list_scope import mocked_scope
from django_forest.utils.scope import ScopeManager


class ScopeCacheInvalidationViewTests(TestCase):
    def setUp(self):
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }

    def tearDown(self):
        # reset _registry after each test
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
    def test_post(self, mocked_scope_has_expired, mocked_decode):
        response = self.client.post(reverse('django_forest:scope-cache-invalidation'),
                                    json.dumps({'renderingId': 1}),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_post_no_renderingId(self, mocked_decode):
        response = self.client.post(reverse('django_forest:scope-cache-invalidation'),
                                     json.dumps({}),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 400)

    def test_post_unauthenticated(self):
        response = self.client.post(reverse('django_forest:scope-cache-invalidation'),
                                     json.dumps({}),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)



import json
from unittest import mock

import sys

import pytest
from django.core.exceptions import ObjectDoesNotExist
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.models import Question, Restaurant, Place
from django_forest.tests.resources.views.list.test_list_scope import mocked_scope
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]

@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
@pytest.mark.usefixtures('reset_config_dir_import')
class ResourceDetailViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json',
                'session.json',
                'question.json', 'choice.json',
                'place.json', 'restaurant.json',
                ]

    def setUp(self, *args, **kwargs):
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:detail', kwargs={'resource': 'tests_question', 'pk': '1'})
        self.reverse_url = reverse('django_forest:resources:detail', kwargs={'resource': 'tests_choice', 'pk': '1'})
        self.one_to_one_url = reverse('django_forest:resources:detail', kwargs={'resource': 'tests_restaurant', 'pk': '1'})
        self.bad_url = reverse('django_forest:resources:detail', kwargs={'resource': 'tests_foo', 'pk': '1'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self, *args, **kwargs):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get(self, *args, **kwargs):
        response = self.client.get(self.url)
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'attributes': {
                    'bar': 'what is your favorite color?+bar',
                    'foo': 'what is your favorite color?+foo',
                    'pub_date': '2021-06-02T13:52:53.528000+00:00',
                    'question_text': 'what is your favorite color?'
                },
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {
                        'data': None,
                        'links': {'related': '/forest/tests_question/1/relationships/topic'}
                    }
                },
                'id': 1,
                'links': {
                    'self': '/forest/tests_question/1'
                }
            },
            'links': {
                'self': '/forest/tests_question/1'
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_get_invalid_token(self, *args, **kwargs):
        response = self.client.get(self.url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'Missing required rendering_id'}]})

    def test_get_no_model(self, *args, **kwargs):
        response = self.client.get(self.bad_url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get_scope(self, *args, **kwargs):
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': 'useless-here'
            }
        }
        response = self.client.get(self.url, {
            'timezone': 'Europe/Paris',
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'attributes': {
                    'bar': 'what is your favorite color?+bar',
                    'foo': 'what is your favorite color?+foo',
                    'pub_date': '2021-06-02T13:52:53.528000+00:00',
                    'question_text': 'what is your favorite color?'
                },
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {'data': None,
                              'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                },
                'id': 1,
                'links': {
                    'self': '/forest/tests_question/1'
                }
            },
            'links': {
                'self': '/forest/tests_question/1'
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get_scope_no_record(self, *args, **kwargs):
        ScopeManager.cache = {
            '1': {
                'scopes': {
                    'tests_question': {
                        'scope': {
                            'filter': {
                                'aggregator': 'and',
                                'conditions': [
                                    {
                                        'field': 'question_text',
                                        'operator': 'equal',
                                        'value': 'color'
                                    },
                                ]
                            },
                            'dynamicScopesValues': {}
                        }
                    }
                },
                'fetched_at': 'useless-here'
            }
        }
        response = self.client.get(self.url, {
            'timezone': 'Europe/Paris',
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': "Record does not exist or you don't have the right to query it"}]})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_put(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'question_text': 'What is your favorite animal?',
                }
            }
        }
        response = self.client.put(self.url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'attributes': {
                    'pub_date': '2021-06-02T13:52:53.528000+00:00',
                    'question_text': 'What is your favorite animal?',
                    'bar': 'what is your favorite color?+bar',
                    'foo': 'what is your favorite color?+foo',
                },
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                },
                'id': 1,
                'links': {
                    'self': '/forest/tests_question/1'
                }
            },
            'links': {
                'self': '/forest/tests_question/1'
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_put_invalid_token(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'question_text': 'What is your favorite animal?',
                }
            }
        }
        response = self.client.put(self.url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'Missing required rendering_id'}]})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_put_related_data(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'serves_pizza': False,
                },
                'relationships': {
                    'place': {
                        'data': {
                            'type': 'tests_places',
                            'id': 2,
                        }
                    },
                }
            }
        }
        self.assertEqual(Restaurant.objects.count(), 1)
        self.assertEqual(Place.objects.count(), 2)
        response = self.client.put(self.one_to_one_url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_restaurant',
                'attributes': {
                    'serves_pizza': False,
                    'serves_hot_dogs': True
                },
                'relationships': {
                    'place': {
                        'links': {
                            'related': '/forest/tests_restaurant/2/relationships/place'
                        }
                    },
                    'waiter_set': {
                        'links': {
                            'related': '/forest/tests_restaurant/2/relationships/waiter_set'
                        }
                    }
                },
                'id': 2,
                'links': {
                    'self': '/forest/tests_restaurant/2'
                }
            },
            'links': {
                'self': '/forest/tests_restaurant/2'
            }
        })
        self.assertEqual(Restaurant.objects.count(), 1)
        self.assertEqual(Place.objects.count(), 2)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_put_related_data_none(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {},
                'relationships': {
                    'question': {
                        'data': None
                    },
                }
            }
        }
        response = self.client.put(self.reverse_url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_choice',
                'id': 1,
                'attributes': {
                    'choice_text': 'yes',
                    'votes': 0
                },
                'relationships': {
                    'question': {
                        'links': {
                            'related': '/forest/tests_choice/1/relationships/question'
                        }
                    },
                    'topic': {
                        'links': {
                            'related': '/forest/tests_choice/1/relationships/topic'
                        }
                    }
                },
                'links': {
                    'self': '/forest/tests_choice/1'
                }
            },
            'links': {
                'self': '/forest/tests_choice/1'
            }
        })

    def test_put_related_data_do_not_exist(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'serves_hot_dogs': True,
                },
                'relationships': {
                    'place': {
                        'data': {
                            'type': 'tests_places',
                            'id': 3,
                        }
                    },
                }
            }
        }
        self.assertEqual(Restaurant.objects.count(), 1)
        response = self.client.put(self.one_to_one_url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Instance tests_place with pk 3 does not exists'
                }
            ]
        })
        self.assertEqual(Restaurant.objects.count(), 1)

    def test_put_no_model(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'question_text': 'What is your favorite animal?',
                }
            }
        }
        response = self.client.put(self.bad_url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_put_smart_field(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'foo': 'What is your favorite animal?',
                    'bar': 'What is your favorite animal?',
                }
            }
        }
        response = self.client.put(self.url,
                                   json.dumps(body),
                                   content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'attributes': {
                    'pub_date': '2021-06-02T13:52:53.528000+00:00',
                    'question_text': 'bar+What is your favorite animal?',
                    'bar': 'what is your favorite color?+bar',
                    'foo': 'what is your favorite color?+foo',
                },
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                },
                'id': 1,
                'links': {
                    'self': '/forest/tests_question/1'
                }
            },
            'links': {
                'self': '/forest/tests_question/1'
            }
        })


    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_delete(self, *args, **kwargs):
        self.assertEqual(Question.objects.count(), 3)
        self.assertTrue(Question.objects.get(pk=1))
        response = self.client.delete(self.url)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Question.objects.count(), 2)
        with self.assertRaises(ObjectDoesNotExist):
            Question.objects.get(pk=1)

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_delete_invalid_token(self, *args, **kwargs):
        response = self.client.delete(self.url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'Missing required rendering_id'}]})

    def test_delete_no_model(self, *args, **kwargs):
        response = self.client.delete(self.bad_url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})



import copy
import json
import os
import logging

import django
from django.conf import settings
from django.db import connection
from django.utils.module_loading import autodiscover_modules

from django_forest.utils.schema.apimap_errors import APIMAP_ERRORS
from django_forest.utils.models import Models
from django_forest.utils.type_mapping import get_type
from django_forest.utils.schema.json_api_schema import create_json_api_schema
from django_forest.utils.forest_api_requester import ForestApiRequester
from .definitions import COLLECTION, FIELD
from .validations import handle_validations
from .enums import handle_enums
from .default import handle_default_value

from .version import get_app_version

from .. import get_accessor_name
from ..forest_setting import get_forest_setting

# Get an instance of a logger
logger = logging.getLogger(__name__)


class Schema:
    schema = {
        'collections': [],
        'meta': {
            'liana': 'django-forestadmin',
            'liana_version': get_app_version(),
            'stack': {
                'database_type': connection.vendor,
                'orm_version': django.get_version(),
            }
        }
    }

    # schema to send to Forest Admin Server
    schema_data = None

    @classmethod
    def get_collection(cls, resource):
        collections = [collection for collection in cls.schema['collections'] if collection['name'] == resource]
        if len(collections):
            return collections[0]

        return None

    @staticmethod
    def get_default(obj, definition):
        for key, value in copy.deepcopy(definition).items():
            obj[key] = value if key not in obj else obj[key]

        return obj

    @staticmethod
    def _get_relation_type(many):
        if many:
            return ['Number']
        return 'Number'

    @staticmethod
    def _get_relationship(field):
        if field.one_to_many or field.many_to_many:
            return 'HasMany'
        elif field.one_to_one:
            return 'HasOne'
        return 'BelongsTo'

    @classmethod
    def handle_relation(cls, field, f):
        if field.is_relation:
            # Notice: do not add if not in included/excluded models
            if field.related_model not in Models.list():
                return None

            many = field.one_to_many or field.many_to_many
            f['field'] = get_accessor_name(field)
            f['type'] = cls._get_relation_type(many)
            f['relationship'] = cls._get_relationship(field)
            # Notice: forest-rails always put id on the end (it should not), do we handle polymorphic support?
            f['reference'] = f'{field.related_model._meta.db_table}.{field.target_field.column}'
            f['is_filterable'] = not many
            f['inverse_of'] = get_accessor_name(field.remote_field)
        return f

    @staticmethod
    def get_type(field):
        type = get_type(field)
        if type in ('Integer', 'Float'):
            return 'Number'
        return type

    @classmethod
    def add_fields(cls, model, collection):
        for field in model._meta.get_fields():
            f = cls.get_default({
                'field': field.name,
                'type': cls.get_type(field)
            }, FIELD)
            f = handle_default_value(field, f)
            f = handle_validations(field, f)
            f = handle_enums(field, f)
            f = cls.handle_relation(field, f)

            if f is not None:
                collection['fields'].append(f)

    @classmethod
    def build_schema(cls):
        cls.schema['collections'] = []
        for model in Models.list():
            collection = cls.get_default({'name': model._meta.db_table}, COLLECTION)
            cls.add_fields(model, collection)
            cls.schema['collections'].append(collection)
        return cls.schema

    @staticmethod
    def add_smart_features():
        # Notice: will load all files in <app>/forest folder from client
        autodiscover_modules(get_forest_setting('CONFIG_DIR', 'forest'))

    @classmethod
    def handle_json_api_schema(cls):
        for collection in cls.schema['collections']:
            # Notice: create marshmallow-jsonapi resource for json api serializer
            create_json_api_schema(collection)

    @classmethod
    def handle_schema_file_production(cls, file_path):
        try:
            with open(file_path, 'r') as f:
                data = f.read()
                try:
                    cls.schema_data = json.loads(data)
                except Exception:
                    logger.error('The content of .forestadmin-schema.json file is not a correct JSON.')
                    logger.error('The schema cannot be synchronized with Forest Admin servers.')
        except Exception:
            logger.error('The .forestadmin-schema.json file does not exist.')
            logger.error('The schema cannot be synchronized with Forest Admin servers.')

    @staticmethod
    def get_serialized_collection(collection):
        for index, field in enumerate(collection['fields']):
            fields = list(FIELD.keys()) + ['validations', 'enums']
            collection['fields'][index] = {x: field[x] for x in field if x in fields}
        return collection

    @classmethod
    def handle_schema_file(cls):
        file_path = os.path.join(os.getcwd(), '.forestadmin-schema.json')
        if settings.DEBUG:
            cls.schema_data = copy.deepcopy(cls.schema)
            for index, collection in enumerate(cls.schema_data['collections']):
                cls.schema_data['collections'][index] = cls.get_serialized_collection(collection)

            with open(file_path, 'w') as f:
                f.write(json.dumps(cls.schema_data, indent=2))
        else:
            cls.handle_schema_file_production(file_path)

    @staticmethod
    def get_serialized_collection_relation(collection, rel_type):
        data = []
        included = []
        for rel in collection[rel_type]:
            id = f"{collection['name']}.{rel['name']}"
            data.append({'id': id, 'type': rel_type})
            included.append({
                'id': id,
                'type': rel_type,
                'attributes': rel
            })
        return data, included

    @classmethod
    def get_serialized_schema(cls):
        data = []
        included = []
        for collection in copy.deepcopy(cls.schema_data['collections']):
            actions_data, actions_included = cls.get_serialized_collection_relation(collection, 'actions')
            segments_data, segments_included = cls.get_serialized_collection_relation(collection, 'segments')
            c = {
                'id': collection['name'],
                'type': 'collections',
                'attributes': cls.get_serialized_collection(collection),
                'relationships': {
                    'actions': {
                        'data': actions_data
                    },
                    'segments': {
                        'data': segments_data
                    }
                }
            }
            data.append(c)
            included.extend(actions_included)
            included.extend(segments_included)

        return {
            'data': data,
            'included': included,
            'meta': cls.schema_data['meta']
        }

    @staticmethod
    def handle_data(r):
        try:
            data = r.json()
        except Exception:
            pass
        else:
            if 'warning' in data:
                logger.warning(data['warning'])

    @classmethod
    def handle_status_code(cls, r):
        if r.status_code in (200, 202, 204):
            cls.handle_data(r)
        elif r.status_code in APIMAP_ERRORS.keys():
            getattr(logger, APIMAP_ERRORS[r.status_code]['level'])(APIMAP_ERRORS[r.status_code]['message'])
        else:
            getattr(logger, APIMAP_ERRORS['error']['level'])(APIMAP_ERRORS['error']['message'])

    @classmethod
    def send_apimap(cls):
        disable_auto_schema_apply = get_forest_setting('FOREST_DISABLE_AUTO_SCHEMA_APPLY', False)
        if not disable_auto_schema_apply:
            serialized_schema = cls.get_serialized_schema()
            try:
                r = ForestApiRequester.post('forest/apimaps', serialized_schema)
            except Exception:
                logger.warning('Cannot send the apimap to Forest. Are you online?')
            else:
                cls.handle_status_code(r)



from unittest import mock

from django.test import TestCase
from django.urls import reverse

from django_forest.authentication.oidc.client_manager import OidcClientManager
from django_forest.tests.authentication.oidc.test_configuration_retriever import mocked_config
from django_forest.tests.authentication.oidc.test_dynamic_client_registrator import mocked_client_credentials


class AuthenticationIndexViewTests(TestCase):

    def setUp(self):
        self.retrieve_patcher = mock.patch('django_forest.authentication.oidc.client_manager.retrieve',
                                           return_value=mocked_config)
        self.register_patcher = mock.patch('django_forest.authentication.oidc.client_manager.register',
                                           return_value=mocked_client_credentials)
        self.mocked_retrieve = self.retrieve_patcher.start()
        self.mocked_register = self.register_patcher.start()
        callback_url = 'http://localhost:8000/authentication/callback'
        self.oidc_client = OidcClientManager.get_client_for_callback_url(callback_url)

    def tearDown(self):
        OidcClientManager.client = None
        self.oidc_client = None
        self.retrieve_patcher.stop()
        self.register_patcher.stop()

    def test_post(self):
        response = self.client.post(reverse('django_forest:authentication:index'),
                                    {'renderingId': '1'},
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)
        r = response.json()
        self.assertEqual(r, {
            'authorizationUrl': 'https://api.development.forestadmin.com/oidc/auth?client_id=eyJraWQiOiI2eWthczh5SVMxdHM3dmZZY3JJRTd3aG1Wc2hwbHVKLUlDbU1pRWszYWJ3IiwiYWxnIjoiUlMyNTYifQ.eyJ0b2tlbl9lbmRwb2ludF9hdXRoX21ldGhvZCI6Im5vbmUiLCJyZWRpcmVjdF91cmlzIjpbImh0dHA6Ly9sb2NhbGhvc3Q6ODAwMC9mb3Jlc3QvYXV0aGVudGljYXRpb24vY2FsbGJhY2siXSwicmVnaXN0cmF0aW9uX2VuZHBvaW50IjoiaHR0cHM6Ly9hcGkuZGV2ZWxvcG1lbnQuZm9yZXN0YWRtaW4uY29tL29pZGMvcmVnIiwiYXBwbGljYXRpb25fdHlwZSI6IndlYiIsImdyYW50X3R5cGVzIjpbImF1dGhvcml6YXRpb25fY29kZSJdLCJyZXNwb25zZV90eXBlcyI6WyJjb2RlIl0sImVudmlyb25tZW50X2lkIjoxMDEsImlzcyI6IkZPUkVTVF9BVVRIRU5USUNBVElPTl9TWVNURU0iLCJpYXQiOjE2MjM0Mjc5NTV9.P_mC8eCr36RqRye8YI5zUSGuq4GmpFf8RLvzQmnVMk2HXqzQToPao_RAyvGpYC4JUjFmuTdGI6SSg9JVGIzNod2AXFAxp8qmK8ax_TBKUlg_F_mzrzOxEOaiFkiCQAhavlBhNCx56_CRBmyt_5Uv6L-E4_ezG0A1kacPfOPZaBNSj8FqCRxrsnVZVq60RmY7tRl0DX50ZUuSEvh0zzZLwayBuTYgxLLdg6EvR52dsJHlv0KgNlnHJwKB74cSAl9CM3PVHym9Eg3V3B_BaAbu0eIKmqtJ7_1h1w3-YsH4jAC-czQNgzgO79IiY2s0tTREe4vv-P-__BJ8lIyFxcgzuQ&response_type=code&scope=openid+email+profile&state=%7B%22renderingId%22%3A+1%7D&redirect_uri=http%3A%2F%2Flocalhost%3A8000%2Fforest%2Fauthentication%2Fcallback'
        })

    def test_post_no_rendering(self):
        response = self.client.post(reverse('django_forest:authentication:index'),
                                    {'foo': '1'},
                                    content_type='application/json')
        self.assertEqual(response.status_code, 500)
        r = response.json()
        self.assertEqual(r,
                         {'errors': [{'status': 500, 'detail': 'Authentication request must contain a renderingId'}]})

    def test_post_bad_rendering(self):
        response = self.client.post(reverse('django_forest:authentication:index'),
                                    {'renderingId': ['foo']},
                                    content_type='application/json')
        self.assertEqual(response.status_code, 500)
        r = response.json()
        self.assertEqual(r,
                         {'errors': [{'status': 500, 'detail': 'The parameter renderingId is not valid'}]})



from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


class JsonApiSerializerMixin:
    def get_pk_name(self, Model):
        if Model._meta.pk.name != 'id':
            return 'pk'
        return 'id'

    def compute_only(self, only, field_name, params, Model):
        collection = Schema.get_collection(Model._meta.db_table)
        field = next((x for x in collection['fields'] if x['field'] == field_name), None)
        if field is not None:
            if field['is_virtual']:
                pk_name = field['reference'].split('.')[1]
                only.append(f'{field_name}.{pk_name}')
                only.remove(field_name)
            else:
                only += [f'{field_name}.{y}' for y in params[f'fields[{field_name}]'].split(',')]
                pk_name = self.get_pk_name(Model._meta.get_field(field_name).related_model)
                only.append(f'{field_name}.{pk_name}')
                only.remove(field_name)
        return only

    def get_deep_only(self, params, Model):
        only = params[f'fields[{Model._meta.db_table}]'].split(',')
        sub_only = [x for x in only if f'fields[{x}]' in params]
        for field_name in sub_only:
            only = self.compute_only(only, field_name, params, Model)
        return only

    def get_smart_relationships(self, resource):
        collection = Schema.get_collection(resource)
        return [x['field'] for x in collection['fields']
                if x['is_virtual'] and x['reference'] and not isinstance(x['type'], list)]

    def get_only(self, params, Model):
        only = []
        lookup = f'fields[{Model._meta.db_table}]'
        if lookup in params:
            only = self.get_deep_only(params, Model)
        if 'context[field]' in params:
            fields_and_relationships = [x.name for x in Model._meta.get_fields()
                                        if not x.is_relation or (x.many_to_one or x.one_to_one)]
            smart_relationships = self.get_smart_relationships(Model._meta.db_table)
            only = fields_and_relationships + smart_relationships

        if 'id' not in only:
            pk_name = self.get_pk_name(Model)
            only.append(pk_name)

        return only

    def get_include_data(self, Model):
        relationships = [x.name for x in Model._meta.get_fields() if x.is_relation and (x.many_to_one or x.one_to_one)]
        smart_relationships = self.get_smart_relationships(Model._meta.db_table)
        return relationships + smart_relationships

    def serialize(self, queryset, Model, params):
        Schema = JsonApiSchema._registry[f'{Model._meta.db_table}Schema']

        kwargs = {
            'include_data': self.get_include_data(Model)
        }

        if f'fields[{Model._meta.db_table}]' in params or 'context[field]' in params:
            kwargs['only'] = self.get_only(params, Model)

        if not queryset:
            data = {'data': []}
        else:
            data = Schema(**kwargs).dump(queryset, many=True)

        return data



from django_forest.utils.collection import Collection
from django_forest.utils.schema import Schema


class SmartFieldMixin:
    def _handle_get_method(self, smart_field, item, resource):
        if 'get' in smart_field:
            method = smart_field['get']
            if isinstance(method, str):
                setattr(item, smart_field['field'], getattr(Collection._registry[resource], method)(item))
            elif callable(method):
                setattr(item, smart_field['field'], method(item))

    def _handle_set_method(self, smart_field, instance, value, resource):
        if 'set' in smart_field:
            method = smart_field['set']
            if isinstance(method, str):
                instance = getattr(Collection._registry[resource], method)(instance, value)
            elif callable(method):
                instance = method(instance, value)
        return instance

    def _add_smart_fields(self, item, smart_fields, resource):
        for smart_field in smart_fields:
            self._handle_get_method(smart_field, item, resource)

    def handle_smart_fields(self, queryset, resource, many=False):
        collection = Schema.get_collection(resource)
        smart_fields = [x for x in collection['fields'] if x['is_virtual']]
        if many:
            for item in queryset:
                self._add_smart_fields(item, smart_fields, resource)
        else:
            self._add_smart_fields(queryset, smart_fields, resource)

    def update_smart_fields(self, instance, body, resource):
        collection = Schema.get_collection(resource)
        smart_fields = [x for x in collection['fields'] if x['is_virtual']]
        for smart_field in smart_fields:
            if smart_field['field'] in body['data']['attributes'].keys():
                value = body['data']['attributes'][smart_field['field']]
                instance = self._handle_set_method(smart_field, instance, value, resource)
        return instance



from django.conf import settings
from django.test import TestCase

from django_forest.utils.middlewares import set_middlewares


class UtilsMiddlewaresTests(TestCase):

    def tearDown(self):
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')

    def test_set_middlewares(self):
        set_middlewares()
        self.assertTrue('django_forest.middleware.PermissionMiddleware' in settings.MIDDLEWARE)



from django_forest.actions.hooks.views.utils import HookView


class ChangeView(HookView):
    def handle_action(self, action, action_name, request):
        body = self.get_body(request.body)
        if self.hook_exists(action, action_name, 'change'):
            changed_field = next((x for x in action['fields'] if x['field'] == body['changedField']), None)
            if changed_field is not None:
                return action['hooks']['change'][changed_field['hook']](body['fields'], request, changed_field)



from urllib.parse import urljoin
from django.urls import reverse

from django_forest.utils.forest_setting import get_forest_setting


def get_callback_url():
    application_url = get_forest_setting('APPLICATION_URL')

    if not application_url:
        raise Exception('APPLICATION_URL is not defined in your FOREST settings')

    return urljoin(application_url, reverse('django_forest:authentication:callback'))



from django.core.management.commands.runserver import Command as BaseRunserverCommand

from django_forest.utils.cors import set_cors
from django_forest.utils.middlewares import set_middlewares
from django_forest.utils.schema import Schema


class Command(BaseRunserverCommand):

    def inner_run(self, *args, **options):
        set_cors()
        set_middlewares()

        # schema
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        Schema.handle_schema_file()
        Schema.send_apimap()
        super(Command, self).inner_run(*args, **options)



import copy
import json
from unittest import mock

from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.models import Restaurant, Publication, Article, Choice
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
class ResourceAssociationListViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json',
                'session.json',
                'question.json', 'choice.json',
                'place.json', 'restaurant.json']

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:associations:list',
                           kwargs={'resource': 'tests_question', 'pk': 1, 'association_resource': 'choice_set'})
        self.reverse_url = reverse('django_forest:resources:associations:list',
                                   kwargs={'resource': 'tests_choice', 'pk': 1, 'association_resource': 'question'})
        self.bad_url = reverse('django_forest:resources:associations:list',
                               kwargs={'resource': 'Foo', 'pk': 1, 'association_resource': 'choice_set'})
        self.bad_association_url = reverse('django_forest:resources:associations:list',
                                           kwargs={'resource': 'tests_question', 'pk': 1, 'association_resource': 'foo'})

        self.many_url = reverse('django_forest:resources:associations:list',
                                kwargs={'resource': 'tests_publication', 'pk': 2, 'association_resource': 'article_set'})
        self.many_reverse_url = reverse('django_forest:resources:associations:list',
                                        kwargs={'resource': 'tests_article', 'pk': 1, 'association_resource': 'publications'})
        self.many_bad_url = reverse('django_forest:resources:associations:list',
                                    kwargs={'resource': 'Foo', 'pk': 2, 'association_resource': 'article_set'})
        self.many_bad_association_url = reverse('django_forest:resources:associations:list',
                                                kwargs={'resource': 'tests_publication', 'pk': 2,
                                                        'association_resource': 'foo'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15',
            'fields[tests_choice]': 'id,question,choice_text',
            'fields[question]': 'question_text'
        })
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_choice',
                    'relationships': {
                        'question': {
                            'links': {
                                'related': '/forest/tests_choice/1/relationships/question'
                            },
                            'data': {
                                'type': 'tests_question',
                                'id': '1'
                            }
                        }
                    },
                    'id': 1,
                    'attributes': {
                        'choice_text': 'yes'
                    },
                    'links': {
                        'self': '/forest/tests_choice/1'
                    }
                },
                {
                    'type': 'tests_choice',
                    'relationships': {
                        'question': {
                            'links': {
                                'related': '/forest/tests_choice/2/relationships/question'
                            },
                            'data': {
                                'type': 'tests_question',
                                'id': '1'
                            }
                        }
                    },
                    'id': 2,
                    'attributes': {
                        'choice_text': 'no'
                    },
                    'links': {
                        'self': '/forest/tests_choice/2'
                    }
                }
            ],
            'included': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                }
            ]
        })

    def test_get_no_model(self, *args, **kwargs):
        response = self.client.get(self.bad_url, {
            'page[number]': '1',
            'page[size]': '15',
            'fields[tests_choice]': 'id,question,choice_text',
            'fields[question]': 'question_text'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource Foo'}]})

    def test_get_no_association(self, *args, **kwargs):
        response = self.client.get(self.bad_association_url, {
            'page[number]': '1',
            'page[size]': '15',
            'fields[tests_choice]': 'id,question,choice_text',
            'fields[question]': 'question_text'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'cannot find association resource foo for Model tests_question'}]})

    def test_post(self, *args, **kwargs):
        publication = Publication.objects.get(pk=2)
        body = {
            'data': [
                {
                    'id': '1',
                    'type': 'article'
                }
            ]
        }
        self.assertEqual(publication.article_set.count(), 1)
        response = self.client.post(self.many_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {})
        self.assertEqual(publication.article_set.count(), 2)

    def test_post_reverse(self, *args, **kwargs):
        article = Article.objects.get(pk=1)
        body = {
            'data': [
                {
                    'id': '2',
                    'type': 'publication'
                }
            ]
        }
        self.assertEqual(article.publications.count(), 1)
        response = self.client.post(self.many_reverse_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {})
        self.assertEqual(article.publications.count(), 2)

    def test_post_no_model(self, *args, **kwargs):
        body = {
            'data': [
                {
                    'id': '1',
                    'type': 'article'
                }
            ]
        }
        response = self.client.post(self.many_bad_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource Foo'}]})

    def test_post_related_data_do_not_exist(self, *args, **kwargs):
        body = {
            'data': [
                {
                    'id': '1',
                    'type': 'article'
                }
            ]
        }
        self.assertEqual(Restaurant.objects.count(), 1)
        response = self.client.post(self.many_bad_association_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'cannot find association resource foo for Model tests_publication'
                }
            ]
        })
        self.assertEqual(Restaurant.objects.count(), 1)

    def test_put(self, *args, **kwargs):
        body = {
            'data': [
                {
                    'id': '1',
                    'type': 'article'
                }
            ]
        }
        response = self.client.put(self.many_url,
                                   json.dumps(body),
                                   content_type='application/json')
        self.assertEqual(response.status_code, 204)

    def test_delete(self, *args, **kwargs):
        data = {
            'data': [
                {
                    'id': '2',
                    'type': 'article'
                }
            ]
        }
        self.assertEqual(Article.objects.count(), 2)
        response = self.client.delete(f'{self.many_url}?delete=true',
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Article.objects.count(), 1)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_delete_all_records(self, *args, **kwargs):
        data = {
            'data': {
                'attributes': {
                    'ids': [
                        {
                            'id': '1',
                            'type': 'choice'
                        }
                    ],
                    'collection_name': 'tests_choice',
                    'parent_collection_name': 'tests_question',
                    'parent_collection_id': '1',
                    'parent_association_name': 'choice_set',
                    'all_records': True,
                    'all_records_subset_query': {
                        'page[number]': 1,
                        'page[size]': 15,
                        'fields[tests_choice]': 'id,question,choice_text,foo,bar'
                    },
                    'all_records_ids_excluded': ['2'],
                    'smart_action_id': None
                },
                'type': 'action-requests'
            }
        }
        self.assertEqual(Choice.objects.count(), 3)
        response = self.client.delete(f'{self.url}?delete=true',
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Choice.objects.count(), 2)

    def test_delete_all_records_no_association(self, *args, **kwargs):
        data = {
            'data': {
                'attributes': {
                    'ids': [
                        {
                            'id': '1',
                            'type': 'choice'
                        }
                    ],
                    'collection_name': 'tests_choice',
                    'parent_collection_name': 'tests_question',
                    'parent_collection_id': '1',
                    'parent_association_name': 'foo',
                    'all_records': True,
                    'all_records_subset_query': {
                        'page[number]': 1,
                        'page[size]': 15,
                        'fields[tests_choice]': 'id,question,choice_text,foo,bar'
                    },
                    'all_records_ids_excluded': ['2'],
                    'smart_action_id': None
                },
                'type': 'action-requests'
            }
        }
        response = self.client.delete(f'{self.url}?delete=true',
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 400)
        data = response.json()
        self.assertEqual(data, {'errors': [{'detail': 'cannot find association resource foo for Model tests_question'}]})

    def test_delete_no_model(self, *args, **kwargs):
        data = {
            'data': [
                {
                    'id': '2',
                    'type': 'article'
                }
            ]
        }
        response = self.client.delete(self.many_bad_url,
                                      json.dumps(data),
                                      content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource Foo'}]})

    def test_delete_no_association(self, *args, **kwargs):
        data = {
            'data': [
                {
                    'id': '2',
                    'type': 'article'
                }
            ]
        }
        response = self.client.delete(self.many_bad_association_url,
                                      json.dumps(data),
                                      content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'cannot find association resource foo for Model tests_publication'}]})

    def test_dissociate(self, *args, **kwargs):
        publication = Publication.objects.get(pk=2)
        data = {
            'data': [
                {
                    'id': '2',
                    'type': 'article'
                }
            ]
        }
        self.assertEqual(publication.article_set.count(), 1)
        self.assertEqual(Article.objects.count(), 2)
        response = self.client.delete(self.many_url,
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(publication.article_set.count(), 0)
        self.assertEqual(Article.objects.count(), 2)



from datetime import timedelta

from dateutil.relativedelta import relativedelta
from django.db.models import Sum, Count

from django_forest.resources.utils.queryset.filters.date.factory import ConditionFactory as DateConditionFactory


def compute_value(params, queryset):
    # sum
    if params['aggregate'] == 'Sum':
        q = queryset.aggregate(sum=Sum(params['aggregate_field']))
        value = q['sum']
        if value is None:
            value = 0
    # count
    else:
        value = queryset.count()

    return int(value)


def get_annotated_queryset(body, queryset, pk_name, association_field_name=None):
    aggregate = body['aggregate']

    # determine name
    name = pk_name  # count by default
    if aggregate == 'Sum':  # sum
        name = body['aggregate_field']

    # do we have an association field?
    if association_field_name is not None:
        name = f'{association_field_name}__{name}'

    # select annotation
    annotation = Count(name)  # count
    if aggregate == 'Sum':
        annotation = Sum(name)  # sum

    return queryset.annotate(annotation), name


def get_format_time_frame(params):
    format = ''
    time_frame = ''
    time_range = params['time_range']
    if time_range == 'Day':
        format = '%d/%m/%Y'
        time_frame = 'days'
    elif time_range == 'Week':
        format = 'W%V-%Y'
        time_frame = 'weeks'
    elif time_range == 'Month':
        format = '%b %y'
        time_frame = 'months'
    elif time_range == 'Year':
        format = '%Y'
        time_frame = 'years'

    return format, time_frame


def get_periods(queryset, key, group_by_date_field, format):
    periods = {}
    for x in queryset:
        tf_formatted = x.get(group_by_date_field).strftime(format)
        if tf_formatted in periods:
            periods[tf_formatted] += x[key]
        else:
            periods[tf_formatted] = x[key]
    return periods


def compute_line_values(bounds, periods, time_frame, format):
    keys = periods.keys()
    values = []
    date = bounds['earliest']
    while date < bounds['latest'] + timedelta(days=1):
        value = 0
        tf_formatted = date.strftime(format)
        if tf_formatted in keys:
            value = periods[tf_formatted]
        values.append({
            'label': tf_formatted,
            'values': {
                'value': value
            }
        })
        date = date + relativedelta(**{time_frame: 1})

    return values


def contains_previous_date_operator(filters):
    if 'aggregator' in filters:
        previous_date_operator = next(
            (x for x in filters['conditions'] if x['operator'] in DateConditionFactory.OFFSET_OPERATORS), None)
        return previous_date_operator is not None
    else:
        return filters['operator'] in DateConditionFactory.OFFSET_OPERATORS



from django_forest.middleware.ip_whitelist import IpWhitelistMiddleware
from django_forest.middleware.permissions import PermissionMiddleware

__all__ = ['PermissionMiddleware', 'IpWhitelistMiddleware']



import json
from unittest import mock

from django.core.serializers.json import DjangoJSONEncoder
from django.test import TestCase, override_settings

from django_forest.utils.forest_api_requester import ForestApiRequester


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code
        self.headers = {'content-type': 'application/json'}
        self.text = json.dumps(json_data, cls=DjangoJSONEncoder)

    def json(self):
        return self.json_data


class MockResponseNoData:
    def __init__(self, status_code):
        self.status_code = status_code


def mocked_requests(json_data, status_code):
    return MockResponse(json_data, status_code)


def mocked_requests_no_data(status_code):
    return MockResponseNoData(status_code)


class UtilsForestApiRequesterTests(TestCase):

    @mock.patch('requests.get', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_get(self, mocked_requests_get):
        r = ForestApiRequester.get('/foo')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json_data, {'key1': 'value1'})

    @override_settings(DEBUG=True)
    @mock.patch('requests.get', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_get_debug(self, mocked_requests_get):
        r = ForestApiRequester.get('/foo')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json_data, {'key1': 'value1'})
        mocked_requests_get.assert_called_once_with(
            'https://api.test.forestadmin.com/foo',
            headers={'Content-Type': 'application/json', 'forest-secret-key': 'foo'},
            params={},
            verify=False
        )

    @mock.patch('requests.get', side_effect=Exception('foo'))
    def test_get_exception(self, mocked_requests_get):
        with self.assertRaises(Exception) as cm:
            ForestApiRequester.get('/foo')

        self.assertEqual(cm.exception.args[0],
                         'Cannot reach Forest API at https://api.test.forestadmin.com/foo, it seems to be down right now.')

    @mock.patch('requests.post', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_post(self, mocked_requests_post):
        r = ForestApiRequester.post('/foo', {'foo': 'bar'})
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json_data, {'key1': 'value1'})

    @override_settings(DEBUG=True)
    @mock.patch('requests.post', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_post_debug(self, mocked_requests_post):
        r = ForestApiRequester.post('/foo', {'foo': 'bar'})
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json_data, {'key1': 'value1'})
        mocked_requests_post.assert_called_once_with(
            'https://api.test.forestadmin.com/foo',
            data=json.dumps({'foo': 'bar'}),
            headers={'Content-Type': 'application/json', 'forest-secret-key': 'foo'},
            params={},
            verify=False
        )

    @mock.patch('requests.post', side_effect=Exception('foo'))
    def test_post_exception(self, mocked_requests_post):
        with self.assertRaises(Exception) as cm:
            ForestApiRequester.post('/foo', {'foo': 'bar'})

        self.assertEqual(cm.exception.args[0],
                         'Cannot reach Forest API at https://api.test.forestadmin.com/foo, it seems to be down right now.')

    @mock.patch('requests.post', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_post_ssl(self, mocked_requests_post):
        r = ForestApiRequester.post('/foo', {'foo': 'bar'})
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json_data, {'key1': 'value1'})

    def test_get_url(self):
        url = ForestApiRequester._get_url('/foo')
        self.assertEqual(url, 'https://api.test.forestadmin.com/foo')

    def test_get_url_ssl(self):
        url = ForestApiRequester._get_url('https://foo.com')
        self.assertEqual(url, 'https://foo.com')



import ipaddress

import requests

from django_forest.utils.forest_api_requester import ForestApiRequester


class IpWhitelist:

    fetched = False
    use_ip_whitelist = False
    rules = []

    @classmethod
    def get_rules(cls):
        response = ForestApiRequester.get('/liana/v1/ip-whitelist-rules')
        if response.status_code != requests.codes.ok:
            raise Exception('Unable to retrieve ip whitelist rules')

        data = response.json()
        cls.fetched = True
        cls.use_ip_whitelist = data['data']['attributes']['use_ip_whitelist']
        cls.rules = data['data']['attributes']['rules']

    @staticmethod
    def same_version(ip1, ip2):
        return isinstance(ip1, type(ip2))

    @staticmethod
    def is_both_loopback(ip1, ip2):
        return ip1.is_loopback and ip2.is_loopback

    @classmethod
    def is_ip_match_ip(cls, ip1, ip2):
        if not cls.same_version(ip1, ip2):
            return cls.is_both_loopback(ip1, ip2)

        if ip1 == ip2:
            return True
        else:
            return cls.is_both_loopback(ip1, ip2)

    @classmethod
    def is_ip_match_range(cls, ip, rule):
        ip_minimum = ipaddress.ip_address(rule['ipMinimum'])
        ip_maximum = ipaddress.ip_address(rule['ipMaximum'])
        if not cls.same_version(ip, ip_minimum):
            return False

        return int(ip_minimum) <= int(ip) <= int(ip_maximum)

    @classmethod
    def is_ip_match_subnet(cls, ip, subnet):
        return ip in list(ipaddress.ip_network(subnet).hosts())

    @classmethod
    def is_ip_matches_rule(cls, ip, rule):
        if rule['type'] == 0:
            return cls.is_ip_match_ip(ipaddress.ip_address(ip), ipaddress.ip_address(rule['ip']))
        elif rule['type'] == 1:
            return cls.is_ip_match_range(ipaddress.ip_address(ip), rule)
        elif rule['type'] == 2:
            return cls.is_ip_match_subnet(ipaddress.ip_address(ip), rule['range'])

    @classmethod
    def is_ip_matches_any_rule(cls, ip):
        for rule in cls.rules:
            if cls.is_ip_matches_rule(ip, rule):
                return True
        else:
            return False



from django_forest.utils.collection import Collection
from .filters import FiltersMixin
from .limit_fields import LimitFieldsMixin
from .pagination import PaginationMixin
from .scope import ScopeMixin
from .search import SearchMixin
from django_forest.resources.utils.decorators import DecoratorsMixin


class QuerysetMixin(PaginationMixin, FiltersMixin, SearchMixin, ScopeMixin, DecoratorsMixin, LimitFieldsMixin):

    def filter_queryset(self, queryset, Model, params, request):
        # Notice: first apply scope
        scope_filters = self.get_scope(request, Model)
        if scope_filters is not None:
            queryset = queryset.filter(scope_filters)

        # Notice: then apply filters and search
        PARAMS = {
            'filters': self.get_filters,
            'search': self.get_search
        }
        for name, method in PARAMS.items():
            if name in params and params[name]:
                queryset = queryset.filter(method(params, Model))
        return queryset

    def enhance_queryset(self, queryset, Model, params, request):
        # scopes + filter + search
        queryset = self.filter_queryset(queryset, Model, params, request)

        # sort
        if 'sort' in params:
            queryset = queryset.order_by(params['sort'].replace('.', '__'))

        # segment
        if 'segment' in params:
            collection = Collection._registry[Model._meta.db_table]
            segment = next((x for x in collection.segments if x['name'] == params['segment']), None)
            if segment is not None and 'where' in segment:
                queryset = queryset.filter(segment['where']())

        # limit fields
        queryset = self.handle_limit_fields(params, Model, queryset)

        # pagination
        queryset = self.get_pagination(params, queryset)

        return queryset



import copy
from datetime import datetime
from unittest import mock

import pytest
import pytz
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
class ResourceListFilterViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.reverse_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_choice'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_is(self, mocked_datetime, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"question_text","operator":"equal","value":"what is your favorite color?"}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_aggregator(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"aggregator":"and","conditions":[{"field":"question_text","operator":"equal","value":"what is your favorite color?"},{"field":"id","operator":"equal","value":1}]}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_aggregator_or(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"aggregator":"or","conditions":[{"field":"id","operator":"equal","value":1},{"field":"id","operator":"equal","value":2}]}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_is_not(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"question_text","operator":"not","value":"what is your favorite color?"}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_is_related_data(self, *args, **kwargs):
        response = self.client.get(self.reverse_url, {
            'fields[tests_choice]': 'id,choice_text,question',
            'fields[question]': 'question_text',
            'filters': '{"field":"question:question_text","operator":"contains","value":"color"}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_choice',
                    'relationships': {
                        'question': {
                            'links': {
                                'related': '/forest/tests_choice/1/relationships/question'
                            },
                            'data': {
                                'type': 'tests_question',
                                'id': '1'
                            }
                        }
                    },
                    'attributes': {
                        'choice_text': 'yes'
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_choice/1'
                    }
                },
                {
                    'type': 'tests_choice',
                    'relationships': {
                        'question': {
                            'links': {
                                'related': '/forest/tests_choice/2/relationships/question'
                            },
                            'data': {
                                'type': 'tests_question',
                                'id': '1'
                            }
                        }
                    },
                    'attributes': {
                        'choice_text': 'no'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_choice/2'
                    }
                }
            ],
            'included': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'question_text': 'what is your favorite color?'
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_wrong_operator(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'filters': '{"field":"question_text","operator":"foo","value":"what is your favorite color?"}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Unknown provided operator foo'
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_blank(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'filters': '{"field":"question_text","operator":"blank","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {'data': []})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_blank_date(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'filters': '{"field":"id","operator":"blank","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {'data': []})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_is_present(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"present","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'id': 2,
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'id': 3,
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from . import views

app_name = 'hooks'
urlpatterns = [
    path('/load', csrf_exempt(views.LoadView.as_view()), name='load'),
    path('/change', csrf_exempt(views.ChangeView.as_view()), name='change'),
]



from django.db.models import Q

from django_forest.resources.utils.queryset.filters.date import DatesMixin
from django_forest.resources.utils.queryset.filters.date.factory import ConditionFactory as DateConditionFactory
from django_forest.utils import get_association_field
from django_forest.utils.type_mapping import get_type
from django_forest.utils.collection import Collection
from django_forest.utils.schema import Schema


OPERATORS = {
    'not': '',
    'contains': '__contains',
    'not_contains': '__contains',
    'before': '__lt',
    'less_than': '__lt',
    'after': '__gt',
    'greater_than': '__gt',
    'starts_with': '__startswith',
    'ends_with': '__endswith',
    'not_equal': '',
    'equal': '',
    'includes_all': '__contains',
    'in': '__in',
}


class ConditionsMixin(DatesMixin):
    def get_basic_expression(self, field, operator, value):
        try:
            lookup_field = f"{field}{OPERATORS[operator]}"
        except Exception:
            raise Exception(f'Unknown provided operator {operator}')
        else:
            is_negated = operator.startswith('not')
            kwargs = {lookup_field: value}

            if is_negated:
                return ~Q(**kwargs)

            return Q(**kwargs)

    def handle_blank(self, field_type, field):
        if field_type == 'String':
            return Q(Q(**{f'{field}__isnull': True}) | Q(**{f'{field}__exact': ''}))
        return Q(**{f'{field}__isnull': True})

    def get_expression_smart_field(self, smart_fields, condition, resource):
        operator = condition['operator']
        field = condition['field'].replace(':', '__')
        value = condition['value']

        smart_field = smart_fields[field]
        if 'filter' in smart_field:
            method = smart_field['filter']
            if isinstance(method, str):
                return getattr(Collection._registry[resource], method)(operator, value)
            elif callable(method):
                return method(operator, value)

    def get_expression_field(self, condition, Model, tz):
        operator = condition['operator']
        field = condition['field'].replace(':', '__')
        value = condition['value']

        field_type = self.get_field_type(condition['field'], Model)

        # special case date, blank and present
        if operator in DateConditionFactory.OPERATORS:
            return self.handle_date_operator(operator, field, value, tz)
        if operator == 'blank':
            return self.handle_blank(field_type, field)
        elif operator == 'present':
            return Q(**{f'{field}__isnull': False})
        else:
            return self.get_basic_expression(field, operator, value)

    def get_expression(self, condition, Model, tz):
        field = condition['field'].replace(':', '__')

        resource = Model._meta.db_table
        collection = Schema.get_collection(resource)
        smart_fields = {x['field']: x for x in collection['fields'] if x['is_virtual']}
        if field in smart_fields.keys():
            return self.get_expression_smart_field(smart_fields, condition, resource)
        else:
            return self.get_expression_field(condition, Model, tz)

    def handle_aggregator(self, filters, Model, tz):
        q_objects = Q()
        for condition in filters['conditions']:
            if filters['aggregator'] == 'or':
                q_objects |= self.get_expression(condition, Model, tz)
            else:
                q_objects &= self.get_expression(condition, Model, tz)
        return q_objects

    def get_field_type(self, field, Model):
        if ':' in field:
            fields = field.split(':')
            RelatedModel = get_association_field(Model, fields[0]).related_model
            field_type = get_type(RelatedModel._meta.get_field(fields[1]))
        else:
            field_type = get_type(Model._meta.get_field(field))

        return field_type



import json
from urllib.parse import urljoin

import requests
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from django_forest.utils.forest_setting import get_forest_setting


class ForestApiRequester:

    @staticmethod
    def get_headers(headers):
        forest_secret_key = get_forest_setting('FOREST_ENV_SECRET')
        base_headers = {
            'Content-Type': 'application/json',
            'forest-secret-key': forest_secret_key,
        }
        base_headers.update(headers)
        return base_headers

    @staticmethod
    def error_msg(url):
        return f'Cannot reach Forest API at {url}, it seems to be down right now.'

    @staticmethod
    def _get_url(route):
        validate = URLValidator()

        try:
            validate(route)
        except ValidationError:
            forest_api_url = get_forest_setting('FOREST_URL', 'https://api.forestadmin.com')
            url = urljoin(forest_api_url, route)
        else:
            url = route
        return url

    @staticmethod
    def get_params(body=None, query=None, headers=None):
        if body is None:
            body = {}
        if query is None:
            query = {}
        if headers is None:
            headers = {}

        return body, query, headers

    @classmethod
    def run_method(cls, method, url, kwargs):
        try:
            r = method(url, **kwargs)
        except Exception:
            raise Exception(cls.error_msg(url))
        else:
            return r

    @classmethod
    def get_from_rendering_id(cls, route, rendering_id):
        query = {
            'renderingId': rendering_id
        }

        response = ForestApiRequester.get(route, query=query)
        if response.status_code == requests.codes.ok:
            return response.json()
        else:
            raise Exception(f'Forest API returned an #{response.content}')

    @classmethod
    def get(cls, route, query=None, headers=None):
        body, query, headers = cls.get_params(query=query, headers=headers)
        url = cls._get_url(route)

        kwargs = {
            'params': query,
            'headers': cls.get_headers(headers)
        }
        if settings.DEBUG:
            kwargs['verify'] = False

        return cls.run_method(requests.get, url, kwargs)

    @classmethod
    def post(cls, route, body=None, query=None, headers=None):
        body, query, headers = cls.get_params(body=body, query=query, headers=headers)
        url = cls._get_url(route)

        kwargs = {
            'data': json.dumps(body),
            'params': query,
            'headers': cls.get_headers(headers)
        }
        if settings.DEBUG:
            kwargs['verify'] = False

        return cls.run_method(requests.post, url, kwargs)



import json

from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.forest_setting import get_forest_setting


def register(metadata):
    initial_access_token = get_forest_setting('FOREST_ENV_SECRET')
    headers = {}
    if initial_access_token:
        headers = {'Authorization': f'Bearer {initial_access_token}'}
    response = ForestApiRequester.post(metadata['registration_endpoint'], body=metadata, headers=headers)

    if response.status_code != 201:
        data = response.json()
        if 'error' in data:
            raise Exception(data)
        raise Exception(MESSAGES['SERVER_TRANSACTION']['REGISTRATION_FAILED'] + json.dumps(data))

    return response.json()



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from . import views

app_name = 'authentication'
urlpatterns = [
    path('', csrf_exempt(views.IndexView.as_view()), name='index'),
    path('/callback', views.CallbackView.as_view(), name='callback'),
    path('/logout', csrf_exempt(views.LogoutView.as_view()), name='logout'),
]



import pandas as pd
from datetime import datetime
try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo


def get_date_range(
    current_datetime: datetime,
    frequency: str,
    periods: int,
    previous: bool = True,
    offset: int = 0,
) -> tuple:
    tzinfo = current_datetime.tzinfo
    current_datetime = current_datetime.replace(tzinfo=zoneinfo.ZoneInfo('UTC'))
    current_dt = pd.to_datetime(current_datetime)
    kwargs = {
        'freq': frequency,
        'periods': offset + 2,
    }
    if periods:
        kwargs['freq'] = f"{periods}{frequency}"

    if previous:
        kwargs['end'] = current_dt
        last_idx = 1
        start_idx = 0
    else:
        kwargs['start'] = current_dt
        last_idx = -1
        start_idx = -2
    date_range = pd.date_range(**kwargs)

    replace_kwargs = {
        'tzinfo': tzinfo,
    }
    if frequency != 'H':
        replace_kwargs = {
            **replace_kwargs,
            'hour': 0,
            'minute': 0,
            'second': 0,
            'microsecond': 0,
        }
    return (
        date_range[start_idx].to_pydatetime().replace(**replace_kwargs),
        date_range[last_idx].to_pydatetime().replace(**replace_kwargs),
    )

def get_previous_x_days(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'd', period, offset=offset)

def get_next_x_days(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'd', period, False, offset=offset)

def get_previous_x_weeks(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'W-MON', period, offset=offset)

def get_previous_x_months(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'MS', period, offset=offset)

def get_previous_x_years(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'YS', period, offset=offset)

def get_previous_x_quarters(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'QS', period, offset=offset)

def get_range_x_hours(current_datetime: datetime, period: int = 0, previous: bool = True, offset: int = 0) -> tuple:
    return get_date_range(current_datetime, 'H', period, previous=previous, offset=offset)

def get_previous_x_hours(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_range_x_hours(current_datetime, period, offset=offset)

def get_next_x_hours(current_datetime: datetime, period: int = 0, offset: int = 0) -> tuple:
    return get_range_x_hours(current_datetime, period, previous=False, offset=offset)



from urllib.parse import urljoin

from oic.oic import Client
from oic.oic.message import ProviderConfigurationResponse

from .configuration_retriever import retrieve
from .dynamic_client_registrator import register


class OidcClientManager:
    client = None

    @classmethod
    def get_client_for_callback_url(cls, callback_url):
        if cls.client:
            return cls.client

        try:
            configuration = retrieve()
            client_credentials = register(
                {
                    'token_endpoint_auth_method': 'none',
                    'redirect_uris': [callback_url],
                    'registration_endpoint': configuration['registration_endpoint']
                })
        except Exception as e:
            raise e
        else:
            client_data = {
                'client_id': client_credentials['client_id'],
                'issuer': configuration['issuer']
            }

            op_info = ProviderConfigurationResponse(
                client_id=client_data['client_id'],
                redirect_uri=callback_url,
                issuer=client_data['issuer'],
                authorization_endpoint=urljoin(client_data['issuer'], 'oidc/auth'),
                token_endpoint=urljoin(client_data['issuer'], 'oidc/token'),
                jwks_uri=urljoin(client_data['issuer'], 'oidc/jwks')
            )

            cls.client = Client(client_data['client_id'], verify_ssl=False)
            cls.client.handle_provider_config(op_info, op_info['issuer'])

            return cls.client



from django.test import TestCase, override_settings


class AuthenticationUtilsTests(TestCase):
    def test_get_callback_url(self):
        from django_forest.authentication.utils import get_callback_url
        callback_url = get_callback_url()
        self.assertEqual(callback_url, 'http://localhost:8000/forest/authentication/callback')

    @override_settings(FOREST={'APPLICATION_URL': None})
    def test_get_callback_url_not_application_url(self):
        from django_forest.authentication.utils import get_callback_url
        with self.assertRaises(Exception) as cm:
            get_callback_url()

        self.assertEqual(cm.exception.args[0], 'APPLICATION_URL is not defined in your FOREST settings')



# Generated by Django 3.2.4 on 2021-07-02 11:47

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0004_serial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='serial',
            name='uuid',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
    ]



from datetime import datetime

from django.http import HttpResponse


class CsvMixin:
    def get_related_res(self, data, value):
        related_res = None
        if value['data'] is not None:
            related_res = next((x for x in data['included']
                                if x['type'] == value['data']['type'] and str(x['id']) == str(value['data']['id'])),
                               None)
        return related_res

    def fill_csv_relationships(self, res, record, data, params):
        for name, value in record['relationships'].items():
            related_res = self.get_related_res(data, value)
            if related_res:
                res[name] = related_res['attributes'][params[f'fields[{name}]']]
        return res

    def fill_csv(self, data, writer, params):
        for record in data['data']:
            res = record['attributes']
            res['id'] = record['id']
            if 'relationships' in record and 'included' in data:
                res = self.fill_csv_relationships(res, record, data, params)

            writer.writerow(res)

    def csv_response(self, csv_filename):
        return HttpResponse(
            content_type='text/csv; charset=utf-8',
            headers={
                'Content-Disposition': f'attachment; filename="{csv_filename}.csv"',
                'Last-Modified': datetime.now(),
                'X-Accel-Buffering': 'no',
                'Cache-Control': 'no-cache'
            },
        )



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from . import views

app_name = 'stats'
urlpatterns = [
    path('', csrf_exempt(views.LiveQueriesView.as_view()), name='liveQueries'),
    path('/<slug:resource>', csrf_exempt(views.StatsWithParametersView.as_view()), name='statsWithParameters'),
]
